-- LOVELACE version 1
-- Copyright JM Alliot 
-- 18 May 2006
-- This is a chess engine xboard compatible written in full ADA
-- with some lines of assembly code
-- It is bitboard based and uses extensively ADA representation clauses
-- Endgames are terrible
-- Draws by repetion are not detected correctly
-- The evaluation function is very bad (and very short..)
-- Use at your own risk...

with Bitboards;
use Bitboards;
with Bitscan;
use Bitscan;
with Text_Io;
use Text_Io;
with Interfaces;
use Interfaces;


package body Eval is
   File_Mask_I : array(0..7) of Intboard:=(others=>0);
   White_King_Bonus : array(0..63) of Integer :=
     (15,15,0,0,0,0,15,15,
      10,10,0,-5,-5,0,10,10,
      -5,-5,-10,-20,-20,-10,-5,-5,
      -20,-20,-20,-20,-20,-20,-20,-20,
      -20,-20,-20,-20,-20,-20,-20,-20,
      -20,-20,-20,-20,-20,-20,-20,-20,
      -20,-20,-20,-20,-20,-20,-20,-20,
      -20,-20,-20,-20,-20,-20,-20,-20);
   Black_King_Bonus : array(0..63) of Integer :=
     (-20,-20,-20,-20,-20,-20,-20,-20,
      -20,-20,-20,-20,-20,-20,-20,-20,
      -20,-20,-20,-20,-20,-20,-20,-20,
      -20,-20,-20,-20,-20,-20,-20,-20,
      -20,-20,-20,-20,-20,-20,-20,-20,
      -5,-5,-10,-20,-20,-10,-5,-5,
      10,10,0,-5,-5,0,10,10,
      15,15,0,0,0,0,15,15);
   King_Files_first : array(0..7) of Integer:=(0,0,0,2,2,5,5,6);
   King_Files_Last : array(0..7) of Integer :=(1,2,2,5,5,7,7,7);
   Black_King_Defenders_I,White_King_Defenders_I : array (0..63) of Intboard := (others=>0);

   function White_King_Safety(Matb:integer) return Integer is
      Pos : Integer;
      Open_White,Open_Black,Open_Both,Bad_Rooks : Integer := 0;
      P_Defenders : Integer;
      coef : Integer;
   begin
      Coef := (-Matb/100-Count_Bits(Black_Pawns_I))-10;
      if Coef<=0 then Return 0; end if;
      Pos := First_One(White_Kings_I);
      P_Defenders := Count_Bits(White_Pawns_I and White_King_Defenders_I(Pos));
      if Pos/8 = 0 then P_Defenders := P_Defenders*2; end if;
      for I in King_Files_First(Pos mod 8)..King_Files_last(Pos mod 8) loop
         if (White_Pawns_I and File_Mask_I(I))=0 then
            Open_White := Open_White+1;
         end if;
         if (Black_Pawns_I and File_Mask_I(I))=0 then
            Open_black := Open_Black+1;
            Bad_Rooks := Count_bits(File_Mask_I(I) and (Black_Rooks_I or Black_Queens_I));
         end if;
         if ((Black_Pawns_I or White_Pawns_I) and File_Mask_I(I)) = 0 then
            Open_Both := Open_Both+1;
            Bad_Rooks := Count_bits(File_Mask_I(I) and (Black_Rooks_I or Black_Queens_I));
         end if;
      end loop;
      return (White_King_Bonus(Pos)-16*(Open_White+Open_Black+2*Open_Both)+8*P_Defenders-16*Bad_Rooks);
   end White_King_Safety;

   function Black_King_Safety(Matw:Integer) return Integer is
      Pos : Integer;
      Open_White,Open_Black,Open_Both,Bad_Rooks : Integer := 0;
      P_Defenders : Integer;
      Coef : Integer;
   begin
      Coef := (Matw/100-Count_Bits(White_Pawns_I))-10;
      if Coef<=0 then Return 0; end if;
      Pos := First_One(Black_Kings_I);
      P_Defenders := Count_Bits(Black_Pawns_I and Black_King_Defenders_I(Pos));
      if Pos/8=7 then P_Defenders := P_Defenders*2; end if;
      for I in King_Files_First(Pos mod 8)..King_Files_last(Pos mod 8) loop
         if (White_Pawns_I and File_Mask_I(I))=0 then
            Open_White := Open_White+1;
            Bad_Rooks := Count_bits(File_Mask_I(I) and (White_Rooks_I or White_Queens_I));
         end if;
         if (Black_Pawns_I and File_Mask_I(I))=0 then
            Open_black := Open_Black+1;
         end if;
         if ((Black_Pawns_I or White_Pawns_I) and File_Mask_I(I)) = 0 then
            Open_Both := Open_Both+1;
            Bad_Rooks := Count_bits(File_Mask_I(I) and (White_Rooks_I or White_Queens_I));
         end if;
      end loop;
      return -(Black_King_Bonus(Pos)-16*(Open_White+Open_Black+2*Open_Both)+8*P_Defenders-16*Bad_Rooks);
   end Black_King_Safety;


   function Piece_Mobility(P : Piece;I : Intboard) return Integer is
      Tmp,From : Integer;
      Mobility : Integer := 0;
      B : Intboard := I;
   begin
      while B/=0 loop
         From := First_One(B);
         B := B xor Set_Mask_I(From);
         if Sliding_Mover(P) then
            if Rook_Mover(P) then
               tmp := Find_Rank(From);
               mobility := Mobility + Rank_Mobility(from,tmp);
               tmp := Find_Rank_90(From);
               Mobility := Mobility + Rank_Mobility_90(from,tmp);
            end if;
            if Bishop_Mover(P) then
               tmp := Find_Rank_45r(From);
               mobility := Mobility + Rank_Mobility_45r(from,tmp);
               tmp := Find_Rank_45l(From);
               mobility := Mobility + Rank_Mobility_45l(from,tmp);
            end if;
         elsif P = Knight then
            Mobility := Knight_Mobility(From);
         end if;
      end loop;
      return Mobility;
   end Piece_Mobility;
   pragma Inline(Piece_Mobility);

   function Valeur(C:Color;Matw:Integer;Matb:Integer;Cast:castling) return Integer is
      Mob : Integer:=0;
      Pos : Integer := 0;
   begin
      Mob := Piece_Mobility(White_Bishop,White_Bishops_I)*2;
      Mob := Mob + Piece_Mobility(White_Rook,White_Rooks_I)*2;
      Mob := Mob + Piece_Mobility(White_Queen,White_Queens_I)/2;
      Mob := Mob + Piece_Mobility(White_Knight,White_Knights_I)*2;

      Mob := Mob - Piece_Mobility(Black_Bishop,Black_Bishops_I);
      Mob := Mob - Piece_Mobility(Black_Rook,Black_Rooks_I)*2;
      Mob := Mob - Piece_Mobility(Black_Queen,Black_Queens_I)/2;
      Mob := Mob - Piece_Mobility(Black_Knight,Black_Knights_I)*2;

      if Cast.Wkd then
         Pos := Pos+30;
      elsif Cast.Wqd then
         Pos := Pos+20;
      elsif (not Cast.Wkc) and then (not Cast.Wqc) then
         Pos := Pos-25;
      end if;
      if Cast.Bkd then
         Pos := Pos-30;
      elsif Cast.Bqd then
         Pos := Pos-20;
      elsif (not Cast.Bkc) and then (not Cast.Bqc) then
         Pos := Pos+25;
      end if;

      return Pos+Mob+Matw+Matb+White_King_Safety(Matb)+Black_King_Safety(Matw);
--      return Pos+Mob+Matw+Matb;
   end Valeur;


   function SEE (C:Color) return Integer is
      Wa,Ba,Wdone,Wtodo,Bdone,Btodo : Intboard;
      From : Integer;

      procedure Test(A : Intboard; D : in out Intboard; I : Intboard ; B : out Boolean;
                                                                       To_Do : out natural) is
         Tmp : Intboard;
      begin
         Tmp := (A and I);
         if  Tmp /= 0 then
            To_Do := First_One(Tmp);
            --               Put_Line("to_do"&Integer'Image(To_Do));
            D := D xor Set_Mask_I(To_do);
            B := True;
         else
            B := False;
         end if;
      end Test;
      pragma Inline(Test);

      function Test_Black return Integer is
         B : Boolean;
         T : Natural;
         Nwa,Nba : Intboard;
      begin
         Test(Btodo,Bdone,Black_Pawns_I,B,T);
         if B then
            Btodo := Ba xor Bdone;
            return Piece_Value(Black_Pawn);
         end if;
         Test(Btodo,Bdone,Black_Knights_I,B,T);
         if B then
            Btodo := Ba xor Bdone;
            return Piece_Value(Black_Knight);
         end if;
         Test(Btodo,Bdone,Black_Bishops_I,B,T);
         if B then
            All_Pieces_45l_I := All_Pieces_45l_I xor Set_Mask_45l_I(T);
            All_Pieces_45r_I := All_Pieces_45r_I xor Set_Mask_45r_I(T);
            All_Diag_Attacks_To(From,Nwa,Nba);
            Wa := Wa or Nwa;
            Ba := Ba or Nba;
            Btodo := Ba xor Bdone;
            return Piece_Value(Black_Bishop);
         end if;
         Test(Btodo,Bdone,Black_Rooks_I,B,T);
         if B then
            All_Pieces_I := All_Pieces_I xor Set_Mask_I(T);
            All_Pieces_90_I := All_Pieces_90_I xor Set_Mask_90_I(T);
            All_Ortho_Attacks_To(From,Nwa,Nba);
            Wa := Wa or Nwa;
            Ba := Ba or Nba;
            Btodo := Ba xor Bdone;
            return Piece_Value(Black_Rook);
         end if;
         Test(Btodo,Bdone,Black_Queens_I,B,T);
         if B then
            All_Pieces_45l_I := All_Pieces_45l_I xor Set_Mask_45l_I(T);
            All_Pieces_45r_I := All_Pieces_45r_I xor Set_Mask_45r_I(T);
            All_Pieces_I := All_Pieces_I xor Set_Mask_I(T);
            All_Pieces_90_I := All_Pieces_90_I xor Set_Mask_90_I(T);
            All_Diag_Attacks_To(From,Nwa,Nba);
            Wa := Wa or Nwa;
            Ba := Ba or Nba;
            All_Ortho_Attacks_To(From,Nwa,Nba);
            Wa := Wa or Nwa;
            Ba := Ba or Nba;
            Btodo := Ba xor Bdone;
            return Piece_Value(Black_Queen);
         end if;
         Test(Btodo,Bdone,Black_Kings_I,B,T);
         if B then
            Btodo := Ba xor Bdone;
            return Piece_Value(Black_King);
         end if;
         raise Constraint_Error;
      end Test_Black;
      pragma Inline(Test_Black);

      function Test_White return Integer is
         B : Boolean;
         T : Natural;
         Nwa,Nba : Intboard;
      begin
         Test(Wtodo,Wdone,White_Pawns_I,B,T);
         if B then
            Wtodo := Wa xor Wdone;
            return Piece_Value(White_Pawn);
         end if;
         Test(Wtodo,Wdone,White_Knights_I,B,T);
         if B then
            Wtodo := Wa xor Wdone;
            return Piece_Value(White_Knight);
         end if;
         Test(Wtodo,Wdone,White_Bishops_I,B,T);
         if B then
            All_Pieces_45l_I := All_Pieces_45l_I xor Set_Mask_45l_I(T);
            All_Pieces_45r_I := All_Pieces_45r_I xor Set_Mask_45r_I(T);
            All_Diag_Attacks_To(From,Nwa,Nba);
            Wa := Wa or Nwa;
            Ba := Ba or Nba;
            Wtodo := Wa xor Wdone;
            return Piece_Value(White_Bishop);
         end if;
         Test(Wtodo,Wdone,White_Rooks_I,B,T);
         if B then
            All_Pieces_I := All_Pieces_I xor Set_Mask_I(T);
            All_Pieces_90_I := All_Pieces_90_I xor Set_Mask_90_I(T);
            All_Ortho_Attacks_To(From,Nwa,Nba);
            Wa := Wa or Nwa;
            Ba := Ba or Nba;
            Wtodo := Wa xor Wdone;
            return Piece_Value(White_Rook);
         end if;
         Test(Wtodo,Wdone,White_Queens_I,B,T);
         if B then
            All_Pieces_45l_I := All_Pieces_45l_I xor Set_Mask_45l_I(T);
            All_Pieces_45r_I := All_Pieces_45r_I xor Set_Mask_45r_I(T);
            All_Pieces_I := All_Pieces_I xor Set_Mask_I(T);
            All_Pieces_90_I := All_Pieces_90_I xor Set_Mask_90_I(T);
            All_Diag_Attacks_To(From,Nwa,Nba);
            Wa := Wa or Nwa;
            Ba := Ba or Nba;
            All_Ortho_Attacks_To(From,Nwa,Nba);
            Wa := Wa or Nwa;
            Ba := Ba or Nba;
            Wtodo := Wa xor Wdone;
            return Piece_Value(White_Queen);
         end if;
         Test(Wtodo,Wdone,White_Kings_I,B,T);
         if B then
            Wtodo := Wa xor Wdone;
            return Piece_Value(White_King);
         end if;
         raise Constraint_Error;
      end Test_White;
      pragma Inline(Test_White);

      W : Intboard := White_Pieces_I;
      B : Intboard := Black_Pieces_I;
      S1,S2,S3,S4 : Intboard;
      V : Integer;
      I : Integer;
      Nvw, Nvb : Integer;
      Values : array(0..32) of Integer;
      Tot_Val : Integer := 0;

   begin
      if C = Black then
         while (W /= 0) loop
            From := First_One(W);
            W := W xor Set_Mask_I (From);
            All_Attacks_To(From,Wa,Ba);
            if Ba /= 0 then
               S1 := All_Pieces_I;
               S2 := All_Pieces_90_I;
               S3 := All_Pieces_45l_I;
               S4 := All_Pieces_45r_I;
               --                  Put_Line("from:"&Integer'Image(From));
               --                  Put_Line("Wa:");
               --                  Print_Intboard(Wa);
               --                  Put_Line("Ba:");
               --                  Print_Intboard(Ba);
               V := 0;
               Btodo := Ba;
               Wtodo := Wa;
               Bdone := 0;
               Wdone := 0;
               Values(0) := V;
               I := 1;
               Nvw := Piece_Value(Chess_Board(From));
               -- Currently only minimax. Should do alpha/beta
               while Btodo /= 0 loop -- No attack, end SEE
                                     --                     Put_Line("Wtodo:");
                                     --                     Print_Intboard(Wtodo);
                                     --                     Put_Line("Btodo");
                                     --                     Print_Intboard(Btodo);
                  Nvb := Test_Black;
                  --                     Put_Line("nvb:"&Integer'Image(Nvb));
                  --Can't take with the king if retake
                  if Wtodo /= 0 and Nvb=Piece_Value(Black_King) then exit; end if;
                  V := V+Nvw;
                  --                     Put_Line("v:"&Integer'Image(v));
                  Values(I) := V;
                  I := I+1;
                  --No attack -> End SEE
                  if Wtodo = 0 then exit; end if;
                  --White's turn
                  --                   Put_Line("Wtodo2:");
                  --                     Print_Intboard(Wtodo);
                  --                     Put_Line("Btodo2:");
                  --                     Print_Intboard(Btodo);
                  Nvw:= Test_White;
                  --                     Put_Line("nvw:"&Integer'Image(Nvw));
                  --Can't take with the king if retake
                  if Btodo /= 0 and Nvw =Piece_Value(White_King) then exit; end if;
                  V := V+Nvb;
                  --                     Put_Line("v:"&Integer'Image(v));
                  Values(I) := V;
                  I := I+1;
               end loop;
               V := Values(I-1);
               for J in reverse 0..I-2 loop
                  if J mod 2 = 0 then -- MAX
                     V := Integer'Max(V,Values(J));
                  else --MIN
                     V := Integer'Min(V,Values(J));
                  end if;
               end loop;
               Tot_Val := Integer'Max(Tot_Val,V);
               All_Pieces_I :=  S1;
               All_Pieces_90_I := S2;
               All_Pieces_45l_I := S3;
               All_Pieces_45r_I := S4;
            end if;
         end loop;
         return Tot_Val;
      else
         while (B /= 0) loop
            From := First_One(B);
            B := B xor Set_Mask_I (From);
            All_Attacks_To(From,Wa,Ba);
            if Wa /= 0 then
               S1 := All_Pieces_I;
               S2 := All_Pieces_90_I;
               S3 := All_Pieces_45l_I;
               S4 := All_Pieces_45r_I;
               --                  Put_Line("from:"&Integer'Image(From));
               --                  Put_Line("Wa:");
               --                  Print_Intboard(Wa);
               --                  Put_Line("Ba:");
               --                  Print_Intboard(Ba);
               V := 0;
               Btodo := Ba;
               Wtodo := Wa;
               Bdone := 0;
               Wdone := 0;
               Values(0) := V;
               I := 1;
               Nvb := Piece_Value(Chess_Board(From));
               -- Currently only minimax. Should do alpha/beta
               while Wtodo /= 0 loop -- No attack, end SEE
                                     --                    Put_Line("Wtodo:");
                                     --                    Print_Intboard(Wtodo);
                                     --                    Put_Line("Btodo");
                                     --                    Print_Intboard(Btodo);
                  Nvw := Test_White;
                  --                    Put_Line("nvw:"&Integer'Image(Nvw));
                  --Can't take with the king if retake
                  if Btodo /= 0 and Nvw>500 then exit; end if;
                  V := V+Nvb;
                  --                     Put_Line("v:"&Integer'Image(v));
                  Values(I) := V;
                  I := I+1;
                  --No attack -> End SEE
                  if Btodo = 0 then exit; end if;
                  --White's turn
                  --                     Put_Line("Wtodo2:");
                  --                     Print_Intboard(Wtodo);
                  --                     Put_Line("Btodo2:");
                  --                     Print_Intboard(Btodo);
                  Nvb:= Test_Black;
                  --                     Put_Line("nvb:"&Integer'Image(Nvb));
                  --Can't take with the king if retake
                  if Wtodo /= 0 and Nvb <-500 then exit; end if;
                  V := V+Nvw;
                  --                     Put_Line("v:"&Integer'Image(v));
                  Values(I) := V;
                  I := I+1;
               end loop;
               V := Values(I-1);
               for J in reverse 0..I-2 loop
                  if J mod 2 = 0 then -- MIN
                     V := Integer'Min(V,Values(J));
                  else --MAX
                     V := Integer'Max(V,Values(J));
                  end if;
               end loop;
               Tot_Val := Integer'Min(Tot_Val,V);
               All_Pieces_I :=  S1;
               All_Pieces_90_I := S2;
               All_Pieces_45l_I := S3;
               All_Pieces_45r_I := S4;
            end if;
         end loop;
         return Tot_Val;
      end if;
   end SEE;

begin
   for I in 0..7 loop
      for J in 0..7 loop
         File_Mask_I(I) := File_Mask_I(I) or Shift_Left(1,8*J+I);
      end loop;
   end loop;
   for I in 0..7 loop
      for J in 0..2 loop
         White_King_Defenders_I(8*J+I) := Shift_Left(1,8*(J+1)+I);
         if (I+1)<8 then
            White_King_Defenders_I(8*J+I) := White_King_Defenders_I(8*J+I) or Shift_Left(1,8*(J+1)+I+1);
         end if;
         if (I-1)>=0 then
            White_King_Defenders_I(8*J+I) := White_King_Defenders_I(8*J+I) or Shift_Left(1,8*(J+1)+I-1);
         end if;
      end loop;
   end loop;
   for I in 0..7 loop
      for J in 6..7 loop
         Black_King_Defenders_I(8*J+I) := Shift_Left(1,8*(J-1)+I);
         if (I+1)<8 then
            Black_King_Defenders_I(8*J+I) := Black_King_Defenders_I(8*J+I) or Shift_Left(1,8*(J-1)+I+1);
         end if;
         if (I-1)>=0 then
            Black_King_Defenders_I(8*J+I) := Black_King_Defenders_I(8*J+I) or Shift_Left(1,8*(J-1)+I-1);
         end if;
      end loop;
   end loop;

end Eval;
