-- LOVELACE version 1
-- Copyright JM Alliot 
-- 18 May 2006
-- This is a chess engine xboard compatible written in full ADA
-- with some lines of assembly code
-- It is bitboard based and uses extensively ADA representation clauses
-- Endgames are terrible
-- Draws by repetion are not detected correctly
-- The evaluation function is very bad (and very short..)
-- Use at your own risk...

with Interfaces;
use Interfaces;
with Adamt19937;
with Bitboards;
use Bitboards;


package body Hash is

   G      : AdaMT19937.Generator;
   Keys   : AdaMT19937.Access_Vector := new AdaMT19937.Vector (0 .. 3);


   Nb_Bits : constant Natural :=20;
   Table : array(0..2**Nb_Bits-1) of Hash_Type
     := (others=>(I=>0,Low=>0,High=>0,From=>0,To=>0,Prof=>-128,Movenum=>0,Valid=>false));

   function Is_In_Table return Boolean is
      Ind : Integer := Integer(Shift_Right(Z_I,64-Nb_Bits));
   begin
      return ((Table(Ind).I=Z_I) and then (Integer(Table(Ind).Prof)/=-128));
   end Is_In_Table;

   -- Warning!!!!!
   -- Only call this function after the above one
   -- You must check that the element is in the table before taking it!
   function Get_From_Table return Hash_Type is
   begin
      return Table(Integer(Shift_Right(Z_I,64-Nb_Bits)));
   end;

   procedure Store_To_Table(Low:Integer;
                            High:Integer;
                            From:Integer;
                            To:Integer;
                            Valid : Boolean;
                            Prof:Integer;
                            Movenum : Integer) is
      Ind : Integer;
   begin
      Ind := Integer(Shift_Right(Z_I,64-Nb_Bits));
      if Z_I = Table(Ind).I then
         -- Exactly same position
         if Prof >= Integer(Table(Ind).Prof) then
            --Same position researched at higher (or equal) depth
            Table(Ind).Low := Integer_16(Low); Table(Ind).High := Integer_16(High);
            Table(Ind).From := Unsigned_6(From); Table(Ind).To := Unsigned_6(To);
            Table(Ind).Prof := Integer_12(Prof); Table(Ind).Movenum := Unsigned_8(Movenum);
            Table(Ind).Valid := Valid;
         else
           -- Same position, lower depth, store should not happen?
           --           raise Constraint_Error;
            null;
         end if;
      else
         -- New position=> collision. Shall we replace the other one?
         if Movenum /= Integer(Table(Ind).Movenum) or else Prof > Integer(Table(Ind).Prof) then
            --store if different move, or better prof
            Table(Ind).I := Z_I;Table(Ind).Valid := Valid;
            Table(Ind).Low := Integer_16(Low); Table(Ind).High := Integer_16(High);
            Table(Ind).From := Unsigned_6(From); Table(Ind).To := Unsigned_6(To);
            Table(Ind).Prof := Integer_12(Prof); Table(Ind).Movenum := Unsigned_8(Movenum);
         end if;
      end if;
   end;

   procedure Make_Z_I(C:Color) is
   begin
      Z_I :=0;
      for I in 0..63 loop
         if Chess_Board(I) /= Empty then
            Z_I := Z_I xor Z_P(I,Chess_Board(I));
         end if;
      end loop;
      if C=Black then Z_I := Z_I xor Z_C; end if;
   end Make_Z_I;

   procedure Check_Hash(C:Color) is
      Tmp : Unsigned_64 := Z_I;
   begin
      Make_Z_I(C);
      if Z_I/=Tmp then raise Constraint_Error; end if;
   end Check_Hash;


begin


   Keys.all := (16#123#, 16#234#, 16#345#, 16#456#);
   AdaMT19937.Reset (G, Keys);

   Z_C := Unsigned_64(AdaMT19937.Random(G)) +
     Shift_Left(Unsigned_64(AdaMT19937.Random(G)),32);
   for I in 0..63 loop
      for J in Piece(0)..Piece(15) loop
         Z_P(I,J) := Unsigned_64(AdaMT19937.Random(G)) +
           Shift_Left(Unsigned_64(AdaMT19937.Random(G)),32);
      end loop;
   end loop;

end Hash;
