-- LOVELACE version 1
-- Copyright JM Alliot 
-- 18 May 2006
-- This is a chess engine xboard compatible written in full ADA
-- with some lines of assembly code
-- It is bitboard based and uses extensively ADA representation clauses
-- Endgames are terrible
-- Draws by repetion are not detected correctly
-- The evaluation function is very bad (and very short..)
-- Use at your own risk...

with Interfaces;
use Interfaces;
with Sequential_Io;
with Bitboards;
use Bitboards;
with Move;
use Move;
with Xboard;
use Xboard;
with Text_Io;
use Text_Io;
with AdaMT19937;
with Calendar;

package body Hashbook is

   type Simple_Move is
      record
         From : Unsigned_8;
         To : Unsigned_8;
      end record;

   type Move_Array is array(0..15) of Simple_Move;

   type Hash_Book is
      record
         Z_I : Unsigned_64;
         Moves : Move_Array;
         Won : Unsigned_16;
         Lost : Unsigned_16;
         Draw : Unsigned_16;
         Next : Natural;
      end record;

   G      : AdaMT19937.Generator;
   Keys   : AdaMT19937.Access_Vector := new AdaMT19937.Vector (0 .. 3);


   Nb_Bits : constant Natural := 16;
   Hash_Index : array(0..2**Nb_Bits-1) of Natural := (others=>0);
   Hash_Tab : array (1..130000) of Hash_Book:=(others=>(Z_I=>0,Moves=>(others=>(From=>255,To=>255)),Won=>0,Lost=>0,Draw=>0,Next=>0));
   Last_Elt : Integer := 1;

   package HTIo is new Sequential_Io(Hash_Book);
   package HIIo is new Sequential_Io(Natural);

   procedure Save_All is
      Fp1 : HTIo.File_Type;
      Fp2 : HIIo.File_Type;
   begin
      HTIo.Create(Fp1,Htio.Out_File,"book.bin");
      for I in Hash_Tab'range loop
         Htio.Write(Fp1,Hash_Tab(I));
      end loop;
      Htio.Close(Fp1);
      HIIo.Create(Fp2,Hiio.Out_File,"booki.bin");
      for I in Hash_Index'Range loop
         Hiio.Write(Fp2,Hash_Index(I));
      end loop;
      Hiio.Close(Fp2);
   end Save_All;

   procedure Load_All is
      Fp1 : HTIo.File_Type;
      Fp2 : HIIo.File_Type;
   begin
      HIIo.Open(Fp2,Hiio.In_File,"booki.bin");
      for I in Hash_Index'Range loop
         Hiio.Read(Fp2,Hash_Index(I));
      end loop;
      Hiio.Close(Fp2);
      HTIo.Open(Fp1,Htio.In_File,"book.bin");
      for I in Hash_Tab'range loop
         Htio.Read(Fp1,Hash_Tab(I));
      end loop;
      Htio.Close(Fp1);
   end Load_All;

   procedure Put_Book(Z_I : Unsigned_64;From,To : Integer;Res : Character) is
      Ind : Integer :=  Integer(Shift_Right(Z_I,64-Nb_Bits));
      T_Ind,Prev : Integer;
   begin
      if Hash_Index(Ind) /= 0 then
         T_Ind := Hash_Index(Ind);
         while (T_Ind /= 0 ) and then (Hash_Tab(T_Ind).Z_I/=Z_I) loop
            Prev := T_Ind;
            T_Ind := Hash_Tab(T_Ind).Next;
         end loop;
         if T_Ind =0 then
            T_Ind := Last_Elt;
            Hash_Tab(Prev).Next := Last_Elt;
            Last_Elt := Last_Elt+1;
         end if;
      else
         T_Ind := Last_Elt;
         Hash_Index(Ind) := T_Ind;
         Last_Elt := Last_Elt+1;
      end if;
      Hash_Tab(T_Ind).Z_I := Z_I;
      if Res = '1' then
         Hash_Tab(T_Ind).won := Hash_Tab(T_Ind).Won+1;
      elsif Res='5' then
         Hash_Tab(T_Ind).draw := Hash_Tab(T_Ind).draw+1;
      else
         Hash_Tab(T_Ind).lost := Hash_Tab(T_Ind).lost+1;
      end if;
      for I in 0..15 loop
         if Hash_Tab(T_Ind).Moves(I).From = Unsigned_8(From) and
           Hash_Tab(T_Ind).Moves(I).To = Unsigned_8(To) then
            exit;
         end if;
         if Hash_Tab(T_Ind).Moves(I).From = 255 then
            Hash_Tab(T_Ind).Moves(I).From := Unsigned_8(From);
            Hash_Tab(T_Ind).Moves(I).To := Unsigned_8(To);
            Exit;
         end if;
      end loop;
   end Put_Book;

   function Get_Ind (Z_I : Unsigned_64) return Integer is
      Ind : Integer :=  Integer(Shift_Right(Z_I,64-Nb_Bits));
      T_Ind : Integer;
   begin
      T_Ind := Hash_Index(Ind);
      while (T_Ind /= 0 ) and then (Hash_Tab(T_Ind).Z_I/=Z_I) loop
         T_Ind := Hash_Tab(T_Ind).Next;
      end loop;
      return T_Ind;
   end Get_Ind;

   procedure Get_Book(From,To : out Integer;Col : Color; Cast : Castling ; En_Passant : Integer) is
      Ind : Integer :=  Integer(Shift_Right(Z_I,64-Nb_Bits));
      T_Ind,N_Ind : Integer;
      St : State;
      Lcb : array (0..63) of Piece;
      P2 : Piece;
      Promote : Boolean;
      N_En_Passant : Integer;
      Valid : Boolean;
      N_Castles : Castling;
      Res : Integer;
      Mv : Integer := -10000;
      Best : Integer:=0;
      T1 : Unsigned_32;
      T2 : Integer;
   begin
      Put_Line(File_Log,"starting get book");
      From := -1; To := -1;
      T_Ind := Get_Ind(Z_I);
      if T_Ind = 0 then return; end if;
      Put_Line(File_Log,"Found something");
      Put_Line(File_Log,Unsigned_16'Image(Hash_Tab(T_Ind).won)&
               Unsigned_16'Image(Hash_Tab(T_Ind).lost)&
               Unsigned_16'Image(Hash_Tab(T_Ind).draw));

      -- Code pour choisir le "meilleur" coup
--      Save_State(St);
--      for J in 0..63 loop Lcb(J):=Chess_Board(J); end loop;
--      for I in Hash_Tab(T_Ind).Moves'Range loop
--         From := Integer(Hash_Tab(T_Ind).Moves(I).From);
--         To := Integer(Hash_Tab(T_Ind).Moves(I).To);
--         if From=255 then exit; end if;
--         Really_Do_Move (From,To,Col,Cast,En_Passant,P2,Promote,N_En_Passant,Valid,N_Castles);
--         N_Ind := Get_Ind(Z_I);
--         if N_Ind /=0 then
--            Res := Integer(Hash_Tab(N_Ind).Won)-Integer(Hash_Tab(N_Ind).Lost);
--         else
--            Res := 0;
--         end if;
--         if Col=Black then Res := -Res; end if;
--         Put_Line(File_Log,Integer'Image(From)&Integer'Image(To)&Integer'Image(Res));
--         if Res>Mv then Mv := Res; Best := I; end if;
--         Restore_State(St);
--         for J in 0..63 loop Chess_Board(J) := Lcb(J); end loop;
--      end loop;

      --code pour choisir un coup au hasard
      T1 := 0;
      for I in Hash_Tab(T_Ind).Moves'Range loop
         if Integer(Hash_Tab(T_Ind).Moves(I).From)=255 then exit; end if;
         T1 := T1+1;
      end loop;
      T2 := Hash_Tab(T_Ind).Moves'First+ Integer(AdaMT19937.Random(G) mod T1);
      Put_Line(File_Log,Unsigned_32'Image(T1)&Integer'Image(T2));
      Flush(File_Log);
      From := Integer(Hash_Tab(T_Ind).Moves(T2).From);
      To := Integer(Hash_Tab(T_Ind).Moves(T2).To);
      Put_Line(File_Log,Integer'Image(From)&Integer'Image(To));
      Flush(File_Log);
   end Get_Book;

   function Get_Last return Integer is
   begin
      return Last_Elt;
   end Get_Last;

   Year,Month,Day : Integer;
   Seconds : Duration;
begin
   Calendar.Split(Calendar.Clock,Year,Month,Day,Seconds);
   Keys.all := (Unsigned_32(Year), Unsigned_32(Month), Unsigned_32(Day), Unsigned_32(Seconds));
   AdaMT19937.Reset (G, Keys);

end Hashbook;
